package test;

import java.util.Collections;
import java.util.Map;
import java.util.ServiceLoader;

import javax.jcr.AccessDeniedException;
import javax.jcr.InvalidItemStateException;
import javax.jcr.ItemExistsException;
import javax.jcr.Node;
import javax.jcr.NodeIterator;
import javax.jcr.PathNotFoundException;
import javax.jcr.PropertyType;
import javax.jcr.ReferentialIntegrityException;
import javax.jcr.Repository;
import javax.jcr.RepositoryException;
import javax.jcr.RepositoryFactory;
import javax.jcr.Session;
import javax.jcr.Value;
import javax.jcr.ValueFormatException;
import javax.jcr.lock.LockException;
import javax.jcr.nodetype.ConstraintViolationException;
import javax.jcr.nodetype.NoSuchNodeTypeException;
import javax.jcr.query.Query;
import javax.jcr.query.QueryManager;
import javax.jcr.query.QueryResult;
import javax.jcr.query.qom.BindVariableValue;
import javax.jcr.query.qom.Column;
import javax.jcr.query.qom.Constraint;
import javax.jcr.query.qom.Join;
import javax.jcr.query.qom.JoinCondition;
import javax.jcr.query.qom.Ordering;
import javax.jcr.query.qom.PropertyValue;
import javax.jcr.query.qom.QueryObjectModel;
import javax.jcr.query.qom.QueryObjectModelFactory;
import javax.jcr.query.qom.Selector;
import javax.jcr.version.VersionException;

import org.modeshape.graph.query.model.SelectorName;
import org.modeshape.jcr.query.qom.JcrEquiJoinCondition;

public class QOMFTest {

	public static void main(String[] args) throws Exception {

		try {
			String configUrl = QOMFTest.class.getResource("/configRepository.xml")
					.toExternalForm();
			Map<String, String> parameters = Collections.singletonMap(
					"org.modeshape.jcr.URL", configUrl);
			Repository repository = null;
			for (RepositoryFactory factory : ServiceLoader
					.load(RepositoryFactory.class)) {
				repository = factory.getRepository(parameters);
				if (repository != null)
					break;
			}

			// Session session = repository.login();

			run(repository);

			System.out.println("Test done");

			// session.logout();

		} finally {
			// System.out.flush();
			// System.exit(0);
		}
	}

	static int counter;
	private static QueryObjectModel createQOM(Session session) throws RepositoryException {
		QueryManager queryManager = session.getWorkspace()
				.getQueryManager();
		QueryObjectModelFactory qomf = queryManager
				.getQOMFactory();
		
		Selector selector1 = qomf.selector("test:test", "a");

// actually propertyName and columnName may be null but NPE is thrown in org.modeshape.graph.query.plan.CanonicalPlanner.attachCriteria
//		Column col = qomf.column(selector1.getSelectorName(), null, null);
//		Column[] columns = new Column[]{col};
		Column[] columns = new Column[]{};
//		Column[] columns = null;
		
		Ordering[] orderings = null;

		Selector selector2 = qomf.selector("test:second", "second");
		
		JoinCondition jc1 = new JcrEquiJoinCondition(new SelectorName(selector1.getSelectorName()), "second", new SelectorName(selector2.getSelectorName()), "jcr:uuid");

		Selector selector3 = qomf.selector("test:owner", "owner");

		JoinCondition jc2 = new JcrEquiJoinCondition(new SelectorName(selector2.getSelectorName()), "owner", new SelectorName(selector3.getSelectorName()), "jcr:uuid");
		
		Join join = qomf.join(selector2, selector3, QueryObjectModelFactory.JCR_JOIN_TYPE_INNER, jc2);
		join = qomf.join(selector1, join, QueryObjectModelFactory.JCR_JOIN_TYPE_INNER, jc1);
		
		BindVariableValue bValue = qomf.bindVariable(selector3.getSelectorName() + "_myAttribute");
		PropertyValue pv = qomf.propertyValue(selector3.getSelectorName(),
				"myAttribute");
		Constraint constraint = qomf.comparison(pv, QueryObjectModelFactory.JCR_OPERATOR_EQUAL_TO, bValue);
		QueryObjectModel qom = qomf.createQuery(
				join, constraint, orderings, columns);
		Value bBound = session.getValueFactory().createValue("again", PropertyType.STRING);
		qom.bindValue(bValue.getBindVariableName(), bBound);
		return qom;
	}
	private static void run(Repository repository) throws RepositoryException,
			ItemExistsException, PathNotFoundException,
			NoSuchNodeTypeException, LockException, VersionException,
			ConstraintViolationException, ValueFormatException,
			AccessDeniedException, ReferentialIntegrityException,
			InvalidItemStateException {

		String id1;

		System.out.println("Test run " + ++counter);

		Session session = repository.login();
		{

			Node root = session.getRootNode();

			Node t1 = root.addNode("firstNode", "test:test");
			Node t2 = root.addNode("secondNode", "test:second");
			Node t3 = root.addNode("ownerNode", "test:owner");
			

			t1.setProperty("second", t2);
			t1.setProperty("myAttribute", "Hello");
			t2.setProperty("owner", t3);
			t2.setProperty("myAttribute", "World");
			t3.setProperty("myAttribute", "again");

			session.save();

			id1 = t1.getIdentifier();

			// session.logout();
		}
//--------------- test with QOMF -----------------------
			// session = repository.login();

			QueryObjectModel qom = createQOM(session);
			QueryResult qr = qom.execute();
			NodeIterator ni = qr.getNodes();
			if(ni.getSize() == 0)
				System.err.println("Hmmm.....");
			while (ni.hasNext()) {
				Node foundNode = (Node) ni.nextNode();
				String primaryNodeType = foundNode.getPrimaryNodeType().getName();
				if(!primaryNodeType.equals("test:test")){
					System.err.println("Found, but wrong nodes: " + primaryNodeType);
				}
				System.err.println(foundNode);
			}

// ----------- lets try with a workaround ;-) ------------
		
		String stmt = qom.getStatement();
		stmt = stmt.replace("*", "a.*");
		stmt = stmt.replace("test:test", "[test:test]");
		stmt = stmt.replace("test:second", "[test:second]");
		stmt = stmt.replace("test:owner", "[test:owner]");
		stmt = stmt.replace("jcr:uuid", "[jcr:uuid]");
		QueryManager queryManager = session.getWorkspace()
				.getQueryManager();
		QueryObjectModelFactory qomf = queryManager
				.getQOMFactory();
		Query q = queryManager.createQuery(stmt, Query.JCR_SQL2);
		BindVariableValue bValue = qomf.bindVariable("owner" + "_myAttribute");
		Value bBound = session.getValueFactory().createValue("again", PropertyType.STRING);
		q.bindValue(bValue.getBindVariableName(), bBound);
		
		QueryResult qr2 = q.execute();
		NodeIterator ni2 = qr2.getNodes();
		if(ni2.getSize() == 0)
			System.err.println("Hmmm.....");
		while (ni2.hasNext()) {
			Node foundNode = (Node) ni2.nextNode();
			String primaryNodeType = foundNode.getPrimaryNodeType().getName();
			if(!primaryNodeType.equals("test:test")){
				System.err.println("Found, but wrong nodes: " + primaryNodeType);
			} else {
				System.out.println("Well done!");
			}
			System.err.println(foundNode);
		}
		session.logout();
		System.exit(0);
	}
}
